#include "common.h"

int board::evalEndgame(int printReport, int alpha, int beta){
	//This function returns an evaluationscore for endgamepositions.
	// To do this, first we calculate a couple of seperate scores:
	// - basic pawnscore
	// - passed pawnscore
	// - hidden pastpawn score
	// - totally past pawnscore
	// - 
	// - 
	// - 
	// - 
	// - 
	//then we merge them into a well weighted value (depending on the gamephase). 
	//We also add a couple of scores that are the result of specific situations if
	//applicable...


    //========================================================================
    //=============== calculate basic pawnscore ==============================
    //========================================================================
	//retrieve basic evalpawnscore, passed pawns and hidden past pawns from hash if 
	//possible. Otherwise, calculate them...  
	BITBOARD pieces;
	BITBOARD ppawns=0;
	BITBOARD hppawns=0;
	int bit;

	int pawnscore=pawnhashtable1->probeHash(whitePawns, blackPawns);
	if (pawnscore==NODATA){
	   //calculate pawn scores...
	   pawnscore=0;
	   //evaluate pawns
	   pieces=whitePawns|blackPawns;
	   while (pieces){
		  bit=findFirstBit(pieces);
		  pawnscore+=evaluatePawn(bit, ((whitePieces>>bit)%2)*2-1, ppawns, hppawns);
		  deleteBit(pieces, bit);      
	   }
	   pawnhashtable1->recordHash(whitePawns, blackPawns, ppawns, hppawns, pawnscore);
	}
	   else{
	   //retrieve ppawns and hppawns from hash...
	   ppawns=pawnhashtable1->probePPawns(whitePawns, blackPawns);
	   hppawns=pawnhashtable1->probeHPPawns(whitePawns, blackPawns);
	}

    //========================================================================
    //====== calculate passed pawnscore & totally ppawnsscore ================
    //========================================================================
	int ppScore=0;
	int tppScore=0;
	if (ppawns){
		//calculate basic scores ppawns
		pieces=ppawns;
		while (pieces){
			  bit=findFirstBit(pieces);
			  ppScore+=ppBonus[(whitePieces>>bit)%2][gamePhase][row(bit)];
			  deleteBit(pieces, bit);      
		}
		//calculate positional scores of passed pawns...
		ppScore+=evalPassedPawns(ppawns);
		//totally passed pawnsscores
		tppScore=evalTotPassedPawns(ppawns);
	}

    //========================================================================
    //=============== calculate hidden passed pawnscore ======================
    //========================================================================
	int hppScore=0; 
	if (hppawns){
		//calculate basic scores hppawns
		pieces=hppawns;
		while (pieces){
			  bit=findFirstBit(pieces);
			  hppScore+=ppBonus[(whitePieces>>bit)%2][gamePhase][row(bit)]/2;
			  deleteBit(pieces, bit);      
		}
		//calculate positional scores of hidden passed pawns...
		hppScore+=evalPassedPawns(ppawns);
	}


	//========================================================================
    //=============== calculate bonus for extra minor pieces =================
    //========================================================================
	int extraMinor=0;
	if ( countBits(whiteBishops|whiteKnights)!=countBits(blackBishops|blackKnights) ){
		int pieceCount[2][6];
		pieceCount[WHITE][QUEEN]=countBits(whiteQueens);
		pieceCount[WHITE][ROOK]=countBits(whiteRooks);
		pieceCount[WHITE][BISHOP]=countBits(whiteBishops);
		pieceCount[WHITE][KNIGHT]=countBits(whiteKnights);
		pieceCount[side(BLACK)][QUEEN]=countBits(blackQueens);
		pieceCount[side(BLACK)][ROOK]=countBits(blackRooks);
		pieceCount[side(BLACK)][BISHOP]=countBits(blackBishops);
		pieceCount[side(BLACK)][KNIGHT]=countBits(blackKnights);

		int majW=pieceCount[WHITE][ROOK] + pieceCount[WHITE][QUEEN];
		int majB=pieceCount[side(BLACK)][ROOK] + pieceCount[side(BLACK)][QUEEN];
		int minW=pieceCount[WHITE][KNIGHT] + pieceCount[WHITE][BISHOP];
		int minB=pieceCount[side(BLACK)][KNIGHT] + pieceCount[side(BLACK)][BISHOP];
		//depending on the number of major and minor pieces for each side a Bonus is granted. 
		if(minW>minB){
			//white has more minors
			if (
				(pieceCount[side(BLACK)][QUEEN]==pieceCount[side(WHITE)][QUEEN]) &&
				(pieceCount[side(BLACK)][ROOK]==pieceCount[side(WHITE)][ROOK]) &&
				whitePawns
				){
				//really an extra minor, Bonus!
				extraMinor+=EXTRA_MINOR*(6-pieceCount[side(BLACK)][QUEEN]-pieceCount[side(BLACK)][ROOK])/6;
			}
			else if (
				(pieceCount[side(BLACK)][QUEEN]==pieceCount[side(WHITE)][QUEEN]) &&
				(pieceCount[side(BLACK)][ROOK]==(pieceCount[side(WHITE)][ROOK]+1) ) &&
				(minW>=(minB+2)) &&
				whitePawns
				){
				//one rook down BUT at least two extra minors, Bonus!
				extraMinor+=EXTRA_MINOR*(6-pieceCount[side(BLACK)][QUEEN]-pieceCount[side(BLACK)][ROOK])/6;
			}
			else if (
				(((pieceCount[side(BLACK)][QUEEN]==(pieceCount[side(WHITE)][QUEEN]+1)) &&
				(pieceCount[side(BLACK)][ROOK]==pieceCount[side(WHITE)][ROOK])) ||//one queen less
				((pieceCount[side(BLACK)][QUEEN]==(pieceCount[side(WHITE)][QUEEN])) &&
				(pieceCount[side(BLACK)][ROOK]==pieceCount[side(WHITE)][ROOK]+2)))//or two rooks less
				&&
				(minW>=(minB+3)) &&
				whitePawns
				){
				//one queen or two rooks down BUT at least three extra minors, Bonus!
				extraMinor+=EXTRA_MINOR*(6-pieceCount[side(BLACK)][QUEEN]-pieceCount[side(BLACK)][ROOK])/6;
			}
		}
		else{
			//black has more minors
			if (
				(pieceCount[side(WHITE)][QUEEN]==pieceCount[side(BLACK)][QUEEN]) &&
				(pieceCount[side(WHITE)][ROOK]==pieceCount[side(BLACK)][ROOK]) &&
				blackPawns
				){
				//really an extra minor, Bonus!
				extraMinor-=EXTRA_MINOR*(6-pieceCount[side(WHITE)][QUEEN]-pieceCount[side(WHITE)][ROOK])/6;
			}
			else if (
				(pieceCount[side(WHITE)][QUEEN]==pieceCount[side(BLACK)][QUEEN]) &&
				(pieceCount[side(WHITE)][ROOK]==(pieceCount[side(BLACK)][ROOK]+1)) &&
				(minB>=(minW+2)) &&
				blackPawns
				){
				//one rook down BUT at least two extra minors, Bonus!
				extraMinor-=EXTRA_MINOR*(6-pieceCount[side(WHITE)][QUEEN]-pieceCount[side(WHITE)][ROOK])/6;
			}
			else if (
				(((pieceCount[side(WHITE)][QUEEN]==(pieceCount[side(BLACK)][QUEEN]+1)) &&
				(pieceCount[side(WHITE)][ROOK]==pieceCount[side(BLACK)][ROOK]))||//one queen less
				((pieceCount[side(WHITE)][QUEEN]==(pieceCount[side(BLACK)][QUEEN])) &&
				(pieceCount[side(WHITE)][ROOK]==pieceCount[side(BLACK)][ROOK]+2)))//or two rooks less
				&&
				(minB>=(minW+3)) &&
				blackPawns
				){
				//one queen or two rooks down BUT at least three extra minors, Bonus!
				extraMinor-=EXTRA_MINOR*(6-pieceCount[side(WHITE)][QUEEN]-pieceCount[side(WHITE)][ROOK])/6;
			}
		}
	}

	int score=lazyEval+pawnscore+ppScore+tppScore+hppScore+extraMinor;

	int margin=100;
	//********************* Try cutoff  *********************************
	if (score*toMove+margin<alpha){
	   return score;
	}
	if (score*toMove-margin>beta){
	   return score;
	}
	// */

    //========================================================================
    //=============== calculate positional score other pieces ================
    //========================================================================
	//also fill the attacks bitboards. They are needed later on.
	BITBOARD attacks[2][6]={0,0,0,0,0,0,0,0,0,0,0,0};

	int piecesPosScore=0;	
	int tropW=0;
	int tropB=0;
	//evaluate knights
	pieces=whiteKnights|blackKnights;
	while (pieces){
		  bit=findFirstBit(pieces);
		  piecesPosScore+=evaluateKnight(bit, ((whitePieces>>bit)%2)*2-1, attacks[(whitePieces>>bit)%2][KNIGHT], tropW, tropB);
		  deleteBit(pieces, bit);      
	}// */

	//evaluate Bishops
	pieces=whiteBishops|blackBishops;
	while (pieces){
		  bit=findFirstBit(pieces);
		  piecesPosScore+=evaluateBishop(bit, ((whitePieces>>bit)%2)*2-1, attacks[(whitePieces>>bit)%2][BISHOP], tropW, tropB);
		  deleteBit(pieces, bit);      
	}// */

	//evaluate Rooks
	pieces=whiteRooks|blackRooks;
	while (pieces){
		  bit=findFirstBit(pieces);
		  piecesPosScore+=evaluateRook(bit, ((whitePieces>>bit)%2)*2-1, attacks[(whitePieces>>bit)%2][ROOK], tropW, tropB);
		  deleteBit(pieces, bit);      
	}// */

	//evaluate Queens
	pieces=whiteQueens|blackQueens;
	while (pieces){
		  bit=findFirstBit(pieces);
		  piecesPosScore+=evaluateQueen(bit, ((whitePieces>>bit)%2)*2-1, attacks[(whitePieces>>bit)%2][QUEEN], tropW, tropB);
		  deleteBit(pieces, bit);      
	}// */

    //========================================================================
    //============ create a territory Bitboard for Black and White ===========
    //========================================================================
    //With this table we calculate which squares could be invaded by the kings
	//when all other pieces would stay where they are. This is very usefull in
	//late endgames. It's score predicts whether a king can infiltrate the
	//opponents camp. The less material is on the board, the more important this
	//score becomes!
    //forbidden are squares the king of that colour cannot come.
    BITBOARD forbidden[2];
    forbidden[WHITE]=	(blackPawns>>7 & ~file[0])|(blackPawns>>9 & ~file[7])| 
						attacks[side(BLACK)][KNIGHT]|
						attacks[side(BLACK)][BISHOP]|
						attacks[side(BLACK)][ROOK]  |
						attacks[side(BLACK)][QUEEN] |
						(whitePieces & ~whiteKings);
    forbidden[side(BLACK)]=	(whitePawns<<7 & ~file[7])|(whitePawns<<9 & ~file[0])| 
						attacks[WHITE][KNIGHT]|
						attacks[WHITE][BISHOP]|
						attacks[WHITE][ROOK]  |
						attacks[WHITE][QUEEN] |
						(blackPieces & ~blackKings);
	//territory is the space that is claimed by this colour
    BITBOARD territory[2];
    territory[WHITE]=whiteKings;
    territory[side(BLACK)]=blackKings;
    //is the territory expanded? notfinished
    int finished[2]={FALSE, FALSE};
    int turn=toMove;
    while(!finished[WHITE] || !finished[side(BLACK)]){
        BITBOARD ter=territory[side(turn)] & //territory of this colour
                     ~territory[side(-turn)] &//we exclude the territory that is 
                                              //also claimed by opponent.
                     ~forbidden[side(turn)]; //own pawns also excluded
                     //note that this excluded territory will be reclaimed
                     //immediately. No new territory however will be claimed
                     //FROM these squares.
        ter=( ((ter<<9|ter<<1|ter>>7) & ~file[0]) | 
              ((ter>>9|ter>>1|ter<<7) & ~file[7]) |   
              (ter|ter>>8|ter<<8 )                  
             );// &  ~forbidden[side(turn)];//own pawns are reclaimed!

        if (ter==territory[side(turn)]){
           //nothing changed
           finished[side(turn)]=TRUE;
        }
        else{
           territory[side(turn)]=ter;
        }
        turn*=-1;                     
    }
    //exclude forbidden terrain
	territory[WHITE]=territory[WHITE]& (~forbidden[WHITE]|whitePieces);
	territory[side(BLACK)]=territory[side(BLACK)] & (~forbidden[side(BLACK)]|blackPieces);
	//printBitboard(forbidden[WHITE]);
	//printBitboard(territory[side(BLACK)]);
    
	//calculate kingInvasionScore
    //penalty for pawns that are in the territory of the opposing king
    //and NOT in the territory of their own monarch
    int kingInvasionScore=0;
	//first white pawns
    BITBOARD pawns=territory[WHITE]& ~territory[side(BLACK)]&blackPawns;
    kingInvasionScore+=VALPAWN * countBits(pawns);
    //and black pawns
    pawns=territory[side(BLACK)]& ~territory[WHITE]& whitePawns;
    kingInvasionScore-=VALPAWN * countBits(pawns);

	if (gamePhase==EARLY_ENDGAME){
		kingInvasionScore*=1/2;
	}


    //========================================================================
    //================ calculate safe passage bonus ppawns====================
    //========================================================================
	//If the path in front of a ppawn is unreachable by the opposing king,
	//this is a clear disadvantage.
    int safePassageScore=0;
	pawns=ppawns & whitePawns;
    while (pawns){
		bit=findFirstBit(pawns);
		//printBitboard(pawnsInFront[WHITE][bit]&~territory[side(BLACK)]);
		//printBitboard(pawnsInFront[WHITE][bit]);
        if ( (pawnsInFront[WHITE][bit]&~territory[side(BLACK)])==pawnsInFront[WHITE][bit]){
                //the path in front of the pawn is completely covered.
                safePassageScore+=VALPAWN/2;
        }
        deleteBit(pawns, bit);
    }
	//the score must be higher if opponent has less majors
	if (safePassageScore && !blackQueens){
		switch(countBits(blackRooks)){
			case 0:
				safePassageScore*=4;
				break;
			case 1:
				safePassageScore*=2;
				break;
		}
	}
	//and black...
	int safePassageBScore=0;
	pawns=ppawns & blackPawns;
    while (pawns){
		bit=findFirstBit(pawns);
        if ( (pawnsInFront[side(BLACK)][bit]&~territory[WHITE])==pawnsInFront[side(BLACK)][bit]){
                //the path in front of the pawn is completely covered.
                safePassageBScore-=VALPAWN/2;
        }
        deleteBit(pawns, bit);
    }
	//the score must be higher if opponent has less majors
	if (safePassageBScore && !whiteQueens){
		switch(countBits(whiteRooks)){
			case 0:
				safePassageBScore*=4;
				break;
			case 1:
				safePassageBScore*=2;
				break;
		}
	}
	safePassageScore+=safePassageBScore;


    //========================================================================
    //======================= calculate evalscore ============================
    //========================================================================
	//of course we start with a lazyEval...
	score=lazyEval+pawnscore+piecesPosScore+ppScore+tppScore+hppScore+
		kingInvasionScore/2+extraMinor+safePassageScore;

	if (printReport){
		printf("\n****************Eval Endgame rapport************************");
		printf("\n**   LazyEval=%d", lazyEval);
		printf("\n**   pawnscore=%d", pawnscore);
		printf("\n**   piecesPosScore=%d", piecesPosScore);
		printf("\n**   ppScore=%d", ppScore);
		printf("\n**   tppScore=%d", (tppScore/3));
		printf("\n**   hppScore=%d", hppScore);
		printf("\n**   kingInvasionScore=%d", kingInvasionScore/2);
		printf("\n**   extraMinor=%d", extraMinor);
		printf("\n**   safePassage=%d", safePassageScore);
		printf("\n**   SCORE=%d", score);
		printf("\n********************end rapport*****************************");
	}

	hashtable1->recordHashEval(boardkey, score);
    return score;
}


