#include "common.h"

int board::evalPawnRookEndgame(int printReport){
	//This function returns an evaluationscore for endgamepositions.
	// To do this, first we calculate a couple of seperate scores:
	// - basic pawnscore
	// - passed pawnscore
	// - hidden pastpawn score
	// - totally past pawnscore


    //========================================================================
    //=============== calculate basic pawnscore ==============================
    //========================================================================
	//retrieve basic evalpawnscore, passed pawns and hidden past pawns from hash if 
	//possible. Otherwise, calculate them...  
	BITBOARD pieces;
	BITBOARD ppawns=0;
	BITBOARD hppawns=0;
	int bit;

	int pawnscore=pawnhashtable1->probeHash(whitePawns, blackPawns);
	if (pawnscore==NODATA){
	   //calculate pawn scores...
	   pawnscore=0;
	   //evaluate pawns
	   pieces=whitePawns|blackPawns;
	   while (pieces){
		  bit=findFirstBit(pieces);
		  pawnscore+=evaluatePawn(bit, ((whitePieces>>bit)%2)*2-1, ppawns, hppawns);
		  deleteBit(pieces, bit);      
	   }
	   pawnhashtable1->recordHash(whitePawns, blackPawns, ppawns, hppawns, pawnscore);
	}
	   else{
	   //retrieve ppawns and hppawns from hash...
	   ppawns=pawnhashtable1->probePPawns(whitePawns, blackPawns);
	   hppawns=pawnhashtable1->probeHPPawns(whitePawns, blackPawns);
	}

 
    //========================================================================
    //===================== calculate positional score rooks =================
    //========================================================================
	BITBOARD attacks[2]={0,0};
	//fill attacks
	pieces=whiteRooks|blackRooks;
	while (pieces){
		  bit=findFirstBit(pieces);
		  attacks[(whitePieces>>bit)%2]|=makeRookMoves(bit);
		  deleteBit(pieces, bit);      
	}// */
	//printBitboard(rookAttacks[1]);


    //========================================================================
    //============ create a territory Bitboard for Black and White ===========
    //========================================================================
    //With this table we calculate which squares could be invaded by the kings
	//when all other pieces would stay where they are. This is very usefull in
	//late endgames. It's score predicts whether a king can infiltrate the
	//opponents camp. The less material is on the board, the more important this
	//score becomes!
	//The attack bitboards contain the squares that are under attack by each side.
	//These squares can NEVER be a part of the kings Territory
	//Pawnattacks to attack Bitboards...
	attacks[WHITE]|=(whitePawns<<7 & ~file[7])|(whitePawns<<9 & ~file[0]);
	attacks[side(BLACK)]|=(blackPawns>>7 & ~file[0])|(blackPawns>>9 & ~file[7]);
	//forbidden are squares that a king cannot claim new territory from. In some cases
	//the king is not allowed to enter this square at all (under attack by opponent or
	//the square occupied by the opposing king). These squares will be removed 
	//from the territory when we are done.
    BITBOARD forbidden[2];
    forbidden[WHITE]= attacks[side(BLACK)]| (allPieces & ~whiteKings);
    forbidden[side(BLACK)]=	attacks[WHITE]| (allPieces & ~blackKings);
	//territory is the space that is claimed by this colour
	//territory is the space that is claimed by this colour
    BITBOARD territory[2];
    territory[WHITE]=whiteKings;
    territory[side(BLACK)]=blackKings;
    //is the territory expanded? notfinished
    int finished[2]={FALSE, FALSE};
    int turn=toMove;
    while(!finished[WHITE] || !finished[side(BLACK)]){
        BITBOARD ter=territory[side(turn)] & //territory of this colour
                     ~territory[side(-turn)] &//we exclude the territory that is 
                                              //also claimed by opponent.
                     ~forbidden[side(turn)]; //own pawns also excluded
                     //note that this excluded territory will be reclaimed
                     //immediately. No new territory however will be claimed
                     //FROM these squares.
        ter=( ((ter<<9|ter<<1|ter>>7) & ~file[0]) | 
              ((ter>>9|ter>>1|ter<<7) & ~file[7]) |   
              (ter|ter>>8|ter<<8 )                  
             );// &  ~forbidden[side(turn)];//own pawns are reclaimed!

        if (ter==territory[side(turn)]){
           //nothing changed
           finished[side(turn)]=TRUE;
        }
        else{
           territory[side(turn)]=ter;
        }
        turn*=-1;                     
    }
    //exclude forbidden terrain
	//Note that the final territory bitboard contains all the squares the king can move to
	//INCLUDING its own pieces as well as the pieces that can be captured by the king!
	territory[WHITE]=territory[WHITE] & ~attacks[side(BLACK)] & ~blackKings;
	territory[side(BLACK)]=territory[side(BLACK)] & ~attacks[WHITE] & ~whiteKings;
	//printBitboard(forbidden[1]);
	//printBitboard(territory[1]);
 

 
	//calculate kingInvasionScore
    //penalty for pawns that are in the territory of the opposing king
    //and NOT in the territory of their own monarch
    int kingInvasionScore=0;
	//first white pawns
    BITBOARD pawns=territory[WHITE]& ~territory[side(BLACK)]&blackPawns;
    kingInvasionScore+=VALPAWN/2 * countBits(pawns);
    //and black pawns
    pawns=territory[side(BLACK)]& ~territory[WHITE]& whitePawns;
    kingInvasionScore-=VALPAWN/2 * countBits(pawns);


    //========================================================================
    //================= calculate passed pawnscore ===========================
    //========================================================================
	int finalppScore=0;
	if (ppawns|hppawns){
		//Which is the most important ppawns? (e.g. the most advanced ppawn)?
		BITBOARD mostImportantPawn=0;
		if(countBits(ppawns)==1){
			mostImportantPawn=ppawns;
		}
		else{
			if(whitePawns&ppawns && blackPawns&ppawns){
				//it is a pawnrace

				//The one that has walked the longest path should be rewarded.
				//determine the distance to queen of the pp's...
				//white
				int QDisW=8;
				BITBOARD bestWhitePawn;
				pawns=ppawns&whitePawns;
				if (pawns){
					int bit=findFirstBit(pawns);
					bestWhitePawn=bit2board[bit];
					QDisW=7-row(bit);
					if (QDisW==6){QDisW--;}
				}
				//black
				int QDisB=8;
				BITBOARD bestBlackPawn;
				pawns=ppawns&blackPawns;
				if (pawns){
					int bit;
					while (pawns){
							bit=findFirstBit(pawns);
							deleteBit(pawns, bit);
					} 
					QDisB=row(bit);
					bestBlackPawn=bit2board[bit];
					if (QDisB==6){QDisB--;}
				}

				//and apply bonus...
				if (QDisB!=QDisW){
					if (QDisB>QDisW && QDisW<5){
						mostImportantPawn=bestWhitePawn;
					}
					if (QDisB<QDisW && QDisB<5){
						mostImportantPawn=bestBlackPawn;
					}
				}// */


			}
		}
		//printBitboard(mostImportantPawn);

		//calculate basic scores ppawns
		pieces=ppawns|hppawns;
		while (pieces){
			bit=findFirstBit(pieces);
			//We start with a basic ppBonus.
			int ppScore=ppBonus[(whitePieces>>bit)%2][gamePhase][row(bit)];
			if(bit2board[bit]&hppawns){
				ppScore/=2;
				if(ppawns){
					//score is even less relevant
					ppScore/=4;
				}
			}
			//now increase/decrease this score depending on many factors.
			if((whitePieces>>bit)%2){
				//First White
				//bonus if the ppawn is defended by a Rook
				if (pawnsInFront[side(BLACK)][bit]& whiteRooks){
				 ppScore+=15;                                                             
				} 
				//penalty if the ppawn is attacked from the front by an opp. rook
				if (pawnsInFront[side(WHITE)][bit]& blackRooks){
				 ppScore-=10;                                                             
				} 
				//bonus if the pawn is flanked by a friendly pawn.
				if (pawnsFlanking[bit] & whitePawns){
					ppScore+=25;
					//Extra bonus if the pawn is flanked by a friendly passed pawn.
					if (pawnsFlanking[bit] & (whitePawns&ppawns)){
						ppScore+=25;
					}
				}
				//Is the path in front of the ppawn NOT in the territory of the opp. king?
				if(!(pawnsInFront[side(WHITE)][bit]&territory[side(BLACK)])){
					//this looks good
					ppScore+=20;
					//is our own monarch in front of the pp (ready to defend it)?
					if(pawnsInFront[side(WHITE)][bit]&whiteKings){
						//very well! Looks like the Lucena position!
						ppScore+=15;
					}
				}
				else{
					//penalty if the opposing king is standing in front of the ppawn
					if (pawnsInFront[side(WHITE)][bit]& blackKings){
						//looks like the philodor position. Reduce score.
						ppScore-=10;                                                             
					} 
				}
			}
			else{
				//now black
				//bonus if the ppawn is defended by a Rook
				if (pawnsInFront[WHITE][bit]& blackRooks){
				 ppScore-=15;                                                             
				} 
				//penalty if the ppawn is attacked from the front by an opp. rook
				if (pawnsInFront[side(BLACK)][bit]& whiteRooks){
				 ppScore+=10;                                                             
				} 
				//bonus if the pawn is flanked by a friendly pawn.
				if (pawnsFlanking[bit] & blackPawns){
					ppScore-=25;
					//Extra bonus if the pawn is flanked by a friendly passed pawn.
					if (pawnsFlanking[bit] & (blackPawns&ppawns)){
						ppScore-=25;
					}
				}
				//Is the path in front of the ppawn NOT in the territory of the opp. king?
				if(!(pawnsInFront[side(BLACK)][bit]&territory[side(WHITE)])){
					//this looks good
					ppScore-=20;
					//is our own monarch in front of the pp (ready to defend it)?
					if(pawnsInFront[side(BLACK)][bit]&blackKings){
						//very well! Looks like the Lucena position!
						ppScore-=15;
					}
				}
				else{
					//penalty if the opposing king is standing in front of the ppawn
					if (pawnsInFront[side(BLACK)][bit]& whiteKings){
						//looks like the philodor position. Reduce score.
						ppScore+=10;                                                             
					} 
				}
			}
			if (!(bit2board[bit]&mostImportantPawn)){
				ppScore=ppScore*3/4;
			}
			finalppScore+=ppScore;
			deleteBit(pieces, bit);      
		}//end this ppawn
	}



    //========================================================================
    //======================= calculate evalscore ============================
    //========================================================================
    int score=lazyEval + pawnscore + kingInvasionScore + finalppScore;

    return score;
}//end of function.

