#include "common.h"

int board::checkMagics(BITBOARD magicNumber, int lengthKey, int square){
	BITBOARD UNUSED=~BITBOARD(0);

	//make table with blockingPieces and fill it
	BITBOARD* blockingPieces=new BITBOARD[1<<12];
	BITBOARD* to=blockingPieces;
	createAllPossibleBoards(to, rookMovesBlockingPieces[square]);

	//calculate number of blocking piece combinations...
	int numPieces=countBits(rookMovesBlockingPieces[square]);
	//and make rookmoves table (clean it first)
	int numBoards=1<<numPieces;
	BITBOARD rookMoves[1<<12];
	for(int i=0; i<numBoards; i++){
		rookMoves[i]=UNUSED;
	}

	//now fill table using the magic number
	for(int i=0; i<numBoards; i++){
		to=(blockingPieces+i);
		BITBOARD rMoves=makeRookMoves(square,*to);
		BITBOARD currentEntry=rookMoves[((*to)*magicNumber)>>(64-lengthKey)];
		if(currentEntry!=UNUSED && currentEntry!=rMoves){
			//printf("!");
			delete blockingPieces;
			return 0;
		}
		rookMoves[((*to)*magicNumber)>>(64-lengthKey)]=rMoves;
	}
	FILE* results=fopen("keys.txt", "a");fprintf(results,"\nvalid key: 0x%I64xULL\n", magicNumber );fclose(results);
	delete blockingPieces;
	return 1;
}


void board::calculateMagicNumbers(){
	int tableBitsUsed[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};
	BITBOARD answer[64];
	int bitsUsed[64];

	for(int i=0; i<64; i++){
		bitsUsed[i]=0;
		printf("int=%d ", i);
		//BITBOARD answer;
		for(int k=0; k<=0; k++){
			k=countBits(rookMovesBlockingPieces[i]);
			BITBOARD fullDestination=(~BITBOARD(0))<<(64-k);
			//printBitboard(fullDestination);
			BITBOARD multiply=0;
			for(int j=0; j<6; j++){
				if(fullDestination&rookMovesBlockingPieces[i]){
					answer[i]=doFirstShift(rookMovesBlockingPieces[i], rookMovesBlockingPieces[i],multiply, fullDestination, j, i);
				}
				else{
					answer[i]=doShift(rookMovesBlockingPieces[i], rookMovesBlockingPieces[i],multiply, fullDestination, j, i);
				}
				if(answer[i]){
					printf("doublecheck=%d   ", checkMagics(answer[i], k, i) );
					printf("bits used=%d ", k );
					tableBitsUsed[k]++;
					bitsUsed[i]=k;
					k=20;
					break;
				}
			}
		}
		printf("key: 0x%I64xULL\n", answer[i] );
	}
	int mem=0;
	for(int i=0; i<13; i++){
		printf("bit=%d   used=%d times=%d\n",tableBitsUsed[i],(1<<i),(tableBitsUsed[i]*(1<<i)) );
		mem+=tableBitsUsed[i]*(1<<i);	
	}
	printf("\nMemory needed=%d\n", mem );
	printf("percentage memory used=%d\n", (mem*100)/102400 );

	//now print output
	printf("\n\nBITBOARD magic_rook[64] = {\n");
	for(int i=0; i<22; i++){
		for(int j=0; j<3 && !(i==22 && j>0); j++){
			printf("0x%I64xULL, ", answer[i*3+j] );
		}
		printf("\n");
	}
	printf("};\n");
	printf("BITBOARD magic_rook_mask[64] = {\n");
	for(int i=0; i<8; i++){
		for(int j=0; j<8; j++){
			printf("%d, ", bitsUsed[i*8+j] );
		}
		printf("\n");
	}
	printf("};\n");


}


BITBOARD board::doFirstShift(BITBOARD &original, BITBOARD processed, BITBOARD multiply, BITBOARD &fullDestination, int dubs, int square){
	//If we come here, there is at least one bit at the start already inn the destination section
	//calculate the location of the first bit 
	int from=findFirstBit(original);
	//we are gonna shift it 0, 1, 2... untill the first bit rolls off the bitboard.
	//e.g. (63-firstBit) times...
	for(int i=from; i<64; i++){
		int shift=63-i;
		//now perform shift  
		BITBOARD shiftedOriginal=original<<shift;
		//update the multiply bitboard:
		BITBOARD newMultiply=multiply|(BITBOARD(1)<<shift);
		//create new board
		BITBOARD newBoard=(shiftedOriginal&fullDestination)|
			              ((shiftedOriginal&~fullDestination)>>shift);// */
		//are all bits in the destination?
		if(!(newBoard&~fullDestination)){
		//if(newBoard==fullDestination){
			//yes it is! This is a potential legal endresult...
			if(checkMagics(newMultiply, countBits(fullDestination), square)){
				return newMultiply;
			}
		}
		else{
			//Iterate into another shift...
			 BITBOARD answer=doShift(original, newBoard, newMultiply, fullDestination, dubs, square);
			 if(answer){
				 return answer;
			 }
		}
	}//end for loop
	return 0;
}


BITBOARD board::doShift(BITBOARD &original, BITBOARD processed, BITBOARD multiply, BITBOARD &fullDestination, int dubs, int square){
	//calculate the bits that are already in the destination slots
	BITBOARD destination=processed&fullDestination;
	//calculate the location of the first bit that still has to be moved
	BITBOARD unmovedBits=processed&~fullDestination;
	int from=findFirstBit(unmovedBits);
	// board temp contains the slots in the destination square we can try to shift to
	BITBOARD temp=~destination&fullDestination;
	while(temp){
		//find first available free space in the destination slots
		//for example:
		//destination slots: 1010----- to=62 
		int to=findFirstBit(temp);
		//calculate the bitnumber that we need to shift
		//for example:
		//shift=62-57=5
		int shift=to-from;
		//now perform shift  
		//for example:
		//source:        0000 00110<<5 =...
		//shiftedSource: 0110----- 
		BITBOARD shiftedOriginal=original<<shift;
		//Now is this shift valid? E.g. Make sure there are no other bits dragged onto an 
		//occupied location in the final board:
		//destination:    1010----- 
		//shiftedSource:  0110----- The second bit can't be put here because that slot is taken already!
		
		int doubles=countBits(destination&shiftedOriginal);
		int newDubs=dubs-doubles;
		//if(!(destination&shiftedOriginal)){
		if(newDubs>=0){
			//This shift is valid!
			//now update the multiply bitboard:
			BITBOARD newMultiply=multiply|(BITBOARD(1)<<shift);
			//create new board
			BITBOARD newBoard=((destination|shiftedOriginal)&fullDestination)|
				              ((shiftedOriginal&~fullDestination)>>shift);// */
			//are all bits in the destination?
			if(!(newBoard&~fullDestination)){
			//if(newBoard==fullDestination){
				//yes it is! This is a legal endresult!
				//return newMultiply;
				if(checkMagics(newMultiply,countBits(fullDestination), square)){
					return newMultiply;
				}

				/*/now try some more possibilities...
				for(int i=0; i<64; i++){
					BITBOARD nm=newMultiply ^ (BITBOARD(1)<<i);
					if(checkMagics(nm,countBits(fullDestination), square)){
						return nm;
					}
				}// */

				/*BITBOARD nm=createAndCheckVariants(newMultiply, countBits(fullDestination), square, 1);
				if(nm){	return nm;}// */


			}
			else{
				//Iterate into another shift...
				 BITBOARD answer=doShift(original, newBoard, newMultiply, fullDestination, newDubs, square);
				 if(answer){
					 return answer;
				 }
			}
		}
		deleteBit(temp, to);
	}//end while loop
	return 0;
}

BITBOARD board::createAndCheckVariants(BITBOARD original, int bitsDestination, int square, int depth){
	if (depth==0){
		if(checkMagics(original,bitsDestination, square)){
			return original;
		}
		else{
			return 0;
		}
	}
	else{
		for(int i=0; i<64; i++){
			BITBOARD nm=original ^ (BITBOARD(1)<<i);
			nm=createAndCheckVariants(nm, bitsDestination, square, depth-1);
			if(nm){	return nm;}
		}
		return 0;
	}
}


BITBOARD board::makeRookMoves(int square, BITBOARD bb){
       int rowGameBoard=(bb>>row(square)*8)%256;
       //horizontal moves...
       BITBOARD moves=(BITBOARD(rook[col(square)][rowGameBoard])<< row(square)*8 );
       //and add vertical moves...
       int fromR90=rotateBitR90[square];
       rowGameBoard=(rotateBitboardR90(bb)>>row(fromR90)*8)%256;//contains all pieces on the board on this row
       rowGameBoard=rook[col(fromR90)][rowGameBoard];//and now the squares it can move to!
       moves |=rotateBBL90[row(fromR90)][rowGameBoard];
       return moves;
}



void board::createAllPossibleBoards(BITBOARD* &solutions, BITBOARD bb){
	int numBits=countBits(bb);
	//store the locations of the bits in an array
	int location[64];
	BITBOARD copy=bb;
	int counter=0;
	while(copy){
		int bit=findFirstBit(copy);
		location[counter]=bit;
		counter++;
		deleteBit(copy, bit);
	}
	//calculate number of solutions
	int numSolutions=1<<numBits;
	//Now calculate all solutions
	for(int i=0; i<numSolutions; i++){
		//transform i into bitboard
		BITBOARD sol=0;
		for(int j=0; j<numBits; j++){
			if((i>>j)%2){
				sol|=BITBOARD(1)<<location[j];
			}
		}
		BITBOARD* to=(solutions+i);
		//printBitboard(sol);
		*(solutions++)=sol;
	}
}