
#include "common.h"

extern FILE* debugfile;


#ifndef searchtreeInit
#define searchtreeInit
hashtable* searchtree::hashtable1;
int searchtree::normalSearchtree=TRUE;
int searchtree::analyze=FALSE;
int searchtree::iterative=TRUE;
EGTBProbe* searchtree::egtb=0;

#endif


MOVE searchtree::think(board &b, int searchTime, int maxdepth){
	bestMoveValue=0;
    //reset historyTable
	kmt.resetHistoryTable();
	kmt.resetKillerMoveTable();
	//set timer
    timer1.setTime(searchTime);
    timer1.startTimer();
    //initiate parameters for speed measurement
    nodes=0;
    total_nodes=0;
	total_time=0;
    int start_time=timer1.get_ms();
    int absolute_start_time=timer1.get_ms();
    int stop_time;
	//reset ply 
	b.ply=0;
    
    //are there enough moves to start a real search?
    movelist ml;
    b.generateLegalMoves(ml);
	b.sortNewMoves(ml, kmt, 0);
	if (ml.moves[0]==0){
       //no legal moves available! 
       if ( b.inCheck(b.toMove) )
          {return CHECKMATE;}
       else
          {return STALEMATE;}
    }
    //only one possible move?   
    if (ml.moves[1]==0){
       //in that case, just return the move!
       return ml.moves[0];
    }
    //fprintf(debugfile," =============BRUTUS===========\n");

    //*************************START SEARCH******************************
    //set castling startSearch Parameter
	b.castleBeginSearch=b.castle;
	//set first value to alpha.
    int alpha=0;
	int eval=b.evaluate(INFINITE, -INFINITE);
    //iterative deepening...
 	int margin=100;
	int startdepth=0;
	if(!iterative){
		startdepth=maxdepth-1;
	}
    for (int depth=startdepth; depth<maxdepth; depth++)
        {
		//set maxDepth...
		maxDepth=depth*2;
        //get first score...       
		board child=b;
		child.makeMove(ml.moves[0]);

		int extended=0;
		//*** check extension ****
		child.positionInCheck=child.inCheck(child.toMove);
		if (child.positionInCheck){
		   extended++;
		}// */

		//*** recap extension ****
		if ( b.recapSquare==((ml.moves[0])>>6)%64 && !extended){
		   extended++;
		}// */


		ml.moveValue[0]=-alphaBeta(ml.moves[0], child, depth+extended, -(alpha+margin), -(alpha-margin));
		if (ml.moveValue[0]>=(alpha+margin) || ml.moveValue[0]<=(alpha-margin)){
			ml.moveValue[0]=-alphaBeta(ml.moves[0], child, depth+extended, -99999, 99999);
		}

        if (timer1.timeIsUp()==TRUE || *searchEnabled==FALSE){
           if (searchEnabled==&normalSearchtree){
              fprintf(debugfile," depth=%d",depth+1);
              fprintf(debugfile,"only first move examined\n");
           }
           break;
        }
        //see if we have a winning move...
        if (ml.moveValue[0]>=INFINITE){
                 if (searchEnabled==&normalSearchtree){
                     fprintf(debugfile," winning move, break!\n");
                     }
                 break;
        }// */


        //cout some info about the first move found
        stop_time=timer1.get_ms();
        if (searchEnabled==&normalSearchtree || analyze){
           //if searchtree is not a pondering tree, print info.
           //otherwise the interface gets too messed up.
		   fprintf(debugfile," %d",depth+1);                    //depth
           fprintf(debugfile,"  %d", ml.moveValue[0]);          //score
           fprintf(debugfile,"  %d", (stop_time - start_time)/10); //time
           fprintf(debugfile,"  %d ", nodes); //nodes
		   printThinking(b, ml.moves[0]);
		   fprintf(debugfile, "\n");
		   start_time=timer1.get_ms();
        }

		//and try to find a better score looking at all other moves...
		int countMoves=1;
		int* nextMove=&ml.moves[1];
		int* nextValue=&ml.moveValue[1];
		int movesSearched=1;
		alpha=ml.moveValue[0];
		while (*nextMove!=0)
			{
			board child=b;
			child.makeMove(*nextMove);
			extended=0;

			//*** check extension ****
			child.positionInCheck=child.inCheck(child.toMove);
			if (child.positionInCheck){
			   extended++;
			}// */

			//*** recap extension ****
			if ( b.recapSquare==((ml.moves[0])>>6)%64 && !extended){
			   extended++;
			}// */

			if(!extended && depth>3  && ((eval+(child.lazyEval-b.lazyEval)*b.toMove)+900)<alpha){
					extended=-1;
			}// */
			else if(
					movesSearched>5								//at least x moves searched fully
					&& depth>2									//at least x nodes from horizon
					&& !((*nextMove>>12)%8)  					//not a promotion move
					&& b.kind[(*nextMove>>6)%64]==EMPTY	//not a capturing move
					&& !extended 
					&& !b.isPassedPawnPash((*nextMove>>6)%64)	//not a passedPawnPush
					//&& 0
					){
						extended=-1;
			}

			//perform iterative search
			*nextValue=-alphaBeta((*nextMove), child, depth+extended, -alpha-1, -alpha);
			//now (re)search if we a reduction and it returned a value higher than alpha
			if(*nextValue>alpha && extended<0){
				*nextValue=-alphaBeta((*nextMove), child, depth, -alpha-1, -alpha);
			}
			//is it higher than alpha and a PV node? Research fully!
			if (*nextValue>alpha){
				*nextValue=-alphaBeta((*nextMove), child, depth, -INFINITE, INFINITE);
				if (*nextValue>alpha){
					alpha=*nextValue;
					if (searchEnabled==&normalSearchtree){
						stop_time=timer1.get_ms();
						fprintf(debugfile," %d",depth+1);                    //depth
						fprintf(debugfile,"  %d", ml.moveValue[0]);          //score
						fprintf(debugfile,"  %d", (stop_time - start_time)/10); //time
						fprintf(debugfile,"  %d", nodes); //nodes
						fprintf(debugfile,"\n");
						start_time=timer1.get_ms();
					}// */
				}
			}

			if (timer1.timeIsUp()==TRUE)
			 {break;}

			countMoves++;   
			nextMove++;
			nextValue++;
			movesSearched++;
        }

        if (timer1.timeIsUp()==TRUE || *searchEnabled==FALSE){
                 if (searchEnabled==&normalSearchtree){
                    fprintf(debugfile," depth=%d",depth+1);
                    fprintf(debugfile,"\n moves searched=%d",countMoves);
                    fprintf(debugfile,"\n");
                    }
                 ml.sortAfterSearch(countMoves);
                 break;
        }
        ml.sortAfterSearch();
        //see if we have a winning move...
        if (ml.moveValue[0]>=INFINITE-100 || ml.moveValue[0]<=-INFINITE+100 ){
				if (searchEnabled==&normalSearchtree || analyze){
				   //if searchtree is not a pondering tree, print info.
				   //otherwise the interface gets too messed up.
                   fprintf(debugfile," depth=%d",depth+1);
				   if(ml.moveValue[0]>=INFINITE){
					   fprintf(debugfile," winning move2, break!: %d\n", ml.moveValue[0]);
				   }
				   else if(ml.moveValue[0]>=0){
					   fprintf(debugfile," Won in: %d ply\n", (INFINITE-ml.moveValue[0]));
				   }

				   else{
					   fprintf(debugfile," Lost in: %d ply\n", (INFINITE+ml.moveValue[0]));
				   }
				}
                break;
        }
        //see if we have all losing moves...
        if (ml.moveValue[0]==-INFINITE){
                 if (searchEnabled==&normalSearchtree){
                    fprintf(debugfile," depth=%d",depth+1);
                    fprintf(debugfile," all losing moves, break!\n");
                    }
                 break;
        }

		kmt.resetKillerMoveTable();
        start_time=timer1.get_ms(); 
        //b.sortAfterSearch(ml, kmt);

    }
    //***************************END OF SEARCH*******************************
    //if you come here: either time is up or otherwise the maxdepth has been
    //reached. The search has been terminated.      

	//cout some info about the last searchround
    if (searchEnabled==&normalSearchtree){
       total_nodes+=nodes;
       fprintf(debugfile,"totalNodes=%d",nodes, "\n");//total_nodes,"\n");
       stop_time=timer1.get_ms();
       if ((stop_time - absolute_start_time)>0){
          fprintf(debugfile,"average speed= %d",(total_nodes /(stop_time - absolute_start_time)));
          fprintf(debugfile,".kNPS ");
          fprintf(debugfile,"total time elapsed= %d\n",(stop_time - absolute_start_time));
       }
 	   fprintf(debugfile," bestmove=");
       ml.printMoveToFile(ml.moves[0]);
       fprintf(debugfile,"\n");
    }// */

	total_time=(stop_time - absolute_start_time);
	bestMoveValue=ml.moveValue[0];
	return ml.moves[0];// */
}


int searchtree::alphaBeta(int move, board &parent, int depth, int alpha, int beta)
{
    //First, get rid of the quiescense nodes.
	if (depth<1){
       return quiescence(move, parent, alpha, beta);
    }

	//if time is up: return immediately!
	if (timer1.timeIsUp()==TRUE || *searchEnabled==FALSE){
		return -INTERRUPTED;
	}

	//******************** enter this node *************************************
	nodes++;

    if (parent.history1.doubleRepetition(parent.absolPos, parent.counter50move)){
       //it is a draw by repeats. Well, Brutus ain't satisfied with that!
       // He wins the game or dies in the attempt! We return a moderate negative 
       //value instead of zero.
       return -parent.toMove*VALPAWN/2;
    }  // */    

    //********************* probe hashtable ************************************
    int val=hashtable1->probeHash(parent.boardkey, depth, alpha, beta);
    if (val!=NODATA){
       return val;
    } // */   int val;


    //************************** probe EGTB ************************************
	if	(
		parent.countBits(parent.allPieces)<=6 &&		//maximum of 6 pieces on the board
		((parent.kind[(move>>6)%64]!=EMPTY)	||			//capturing move...
		parent.ply<3 )									//...or ply is just 1 or 2
		){
		int egtbVal=EMPTY;
		if (egtb->DoEGTBProbe(parent, &egtbVal)) {
			return egtbVal;
		}
	}// */
		
    
    //********************* Nullmove pruning ***********************************
    int incheck=parent.inCheck(parent.toMove);
	int NMP_reduction=3;
	if(depth>6){NMP_reduction=4;}
    if (parent.precedingNullMoves<1 && !incheck){
       board child=parent;
	   child.makeMove(NULLMOVE);
	   val=-alphaBeta(NULLMOVE, child, (depth-NMP_reduction), -beta, -beta+1);
       if (val>=beta){
          return beta;
       }
    }// */


    //*************************************************************************
    //********************** Start full search ********************************
    //*************************************************************************
    int hashf=hashfALPHA;
	int bestMove=EMPTY;
	int hashMove=EMPTY;
	int movesSearched=0;
	movelist ml;
	int eval=parent.evaluate(INFINITE, -INFINITE);

	for (int searchphase=0; searchphase<5; searchphase++){
		if (searchphase==0){
			//****************** Probe hashtable for best move *************************
			//We do this before generating moves.
			hashMove=hashtable1->probeBestMove(parent.boardkey);
			if(hashMove!=EMPTY){
				ml.moves[0]=hashMove;
				ml.moves[1]=0;
			}
			else{ml.moves[0]=0;}
		}
		if (searchphase==1){
			//****************** internal iterative deepening *************************
			if (hashMove==EMPTY && depth>=4 && move!=NULLMOVE){
				//First do a search with a decreased depth...
				int score=alphaBeta(move, parent, (depth-3), alpha, beta);
				//and probe hashtable for best move. If all went well, we just generated a hashMove
				hashMove=hashtable1->probeBestMove(parent.boardkey);
				if(hashMove!=EMPTY){
					ml.moves[0]=hashMove;
					ml.moves[1]=0;
				}
				else{searchphase++;}
			}// 			
			else{searchphase++;}
		}
		if (searchphase==2){
			if(incheck){
				//********************** Generate evasive moves ***************************
				parent.generateSortedEvasiveMoves(ml, kmt);
				searchphase=INFINITE;
			}
			else{
				//********************** Generate capmoves ********************************
				//no move found in hash or it wasn't good enough. Generate capmoves...   
				parent.generateSortedCaptureMoves(ml, kmt);
			}
		}
		if (searchphase==3){
			//********************** Generate killermoves *****************************
			parent.generateLegalKillerMoves(ml, kmt);
		}

		if (searchphase==4){
			//********************** Generate moves ***********************************
			//no capmove found in hash or it wasn't good enough. Generate normal noncapturing moves...   
			parent.generateSortedNonCaptureMoves(ml, kmt, depth);
			movesSearched=100;
		}

		//*********************** start deeper search ****************************
		//Find value by -alphaBeta (or if possible by futility cutoff).
		int* nextMove=&ml.moves[0];
		while (*nextMove!=0){

			board child=parent;
			child.makeMove(*nextMove);

			//only for legal moves...
			if(!child.isMoveIllegal(*nextMove)){
				movesSearched++;

				//********************* several extensions *********************************
				int extended=0;

				//*** check extension ****
				child.positionInCheck=child.inCheck(child.toMove);
				if (child.positionInCheck){
				   extended++;
				}// */

				//*** recap extension ****
				if ( parent.recapSquare==((*nextMove)>>6)%64 && !extended){
				   extended++;
				}// */

				//********************* Futility pruning *********************************
				int futPred=9;
				int pruned=FALSE;
				int pruning=9;
				int eval1;
				
				if(!child.positionInCheck && !parent.positionInCheck && depth==1 && ((eval+(child.lazyEval-parent.lazyEval)*parent.toMove)+350)<alpha){
					val=alpha;
					pruned=TRUE;
				} 

				if(!child.positionInCheck && !parent.positionInCheck && depth==2 && ((eval+(child.lazyEval-parent.lazyEval)*parent.toMove)+600)<alpha){//dbg::futMargin
					val=alpha;
					pruned=TRUE;
				}

					/*if (dbg::lmpBench){
						futPred=ALPHA;
					}
				} 
				else if (dbg::lmpBench){
					futPred=BETA;
				}//

				//now benching data
				if(futPred!=9){
					val=-alphaBeta((*nextMove), child, 0, -beta, -alpha);
					if (val<=alpha){
						dbg::lmp[futPred][ALPHA]++;
					}
					else{
						dbg::lmp[futPred][BETA]++;
					}
				}// */

				if(!child.positionInCheck && depth==2  && movesSearched>5 && ((eval+(child.lazyEval-parent.lazyEval)*parent.toMove)+300)<alpha){
					extended--;
				} 
				if(!child.positionInCheck && depth==3  && movesSearched>5 && ((eval+(child.lazyEval-parent.lazyEval)*parent.toMove)+600)<alpha){
					extended--;
				} 
				if(!child.positionInCheck && depth>3  && ((eval+(child.lazyEval-parent.lazyEval)*parent.toMove)+900)<alpha){
					extended--;
				}// */

				//*** negascout algorithm ***
				dbg::total++;
				if(!pruned){
					if (movesSearched==0 || !dbg::futActive){//first move fully searched
						val=-alphaBeta((*nextMove), child, depth-1+extended, -beta, -alpha);			  
						//now benching data
						if(!dbg::futActive && futPred!=9){
							if (val<=alpha){dbg::lmp[futPred][ALPHA]++;}
							else{dbg::lmp[futPred][BETA]++;}
						}// */

					}
					else{
						//try late move pruning
						int statEval=100;
						if(parent.kind[(*nextMove>>6)%64]!=EMPTY){
							statEval=parent.staticEvaluate( ((*nextMove)%64), ((*nextMove>>6)%64) );
						}

						if (												//try late move pruning if:
								//movesSearched>102 &&
								(movesSearched>102 || statEval<0 ) &&		//at least x moves searched fully
								!child.positionInCheck 						//no in check
								&& !parent.positionInCheck 					//no in check
								&& depth>2									//at least x nodes from horizon
								&& (hashtable1->probeBestMove(parent.boardkey)!=EMPTY)
								&& !((*nextMove>>12)%8)  					//not a promotion move
								//&& parent.kind[(*nextMove>>6)%64]==EMPTY	//not a capturing move
								&& kmt.getKillerMoveScore(*nextMove, parent.ply, side(parent.toMove))==0
								&& !extended
								&& !parent.isPassedPawnPash((*nextMove>>6)%64)	//not a passedPawnPush
								//&& !parent.isMoveAgressive(*nextMove)
								//&& 0
								){
									extended=-1;
									if (dbg::lmpBench){futPred=8;}

						}
						//perform iterative search
						val=-alphaBeta((*nextMove), child, depth-1+extended, -alpha-1, -alpha);

						//dbg part 1
						if(futPred!=9){
							if (val<=alpha){
								futPred=ALPHA;
							}
							else{
								futPred=BETA;
							}
							val=alpha+1;
						}

						//now (re)search if we a reduction and it returned a value higher than alpha
						if(val>alpha && extended<0){
							val=-alphaBeta((*nextMove), child, depth-1, -alpha-1, -alpha);
						}

						//dbg part 2
						if(futPred!=9){
							if (val<=alpha){
								dbg::lmp[futPred][ALPHA]++;
							}
							else{
								dbg::lmp[futPred][BETA]++;
							}
						}


						//is it higher than alpha and a PV node? Research fully!
						if(val>alpha && val<beta){
							if(extended<0){extended=0;}
							val=-alphaBeta((*nextMove), child, depth-1+extended, -beta, -alpha);	
						}
					}
				}

				if (val==INTERRUPTED){
				  //time is up! return immediately!
				  return -INTERRUPTED;
				}                  

				if (val>=INFINITE && alpha<INFINITE){
					//move is winning move. No point in continuing the search...
					return val;
				}
				if (val>=beta){
					hashtable1->recordHash(parent.boardkey, depth, val, hashfBETA);
					hashtable1->storeBestMove(parent.boardkey, *nextMove);
					kmt.storeKillerMove(*nextMove, val, parent.ply, side(parent.toMove), parent.kind[((*nextMove)>>6)%64]);
					return beta;
				}
				/*if(parent.kind[(*nextMove>>6)%64]==EMPTY && parent.isMoveAgressive(*nextMove)){
					kmt.storeKillerMove(*nextMove, val, parent.ply, side(parent.toMove), parent.kind[((*nextMove)>>6)%64]);
				}*/


				if (val>alpha){
					bestMove=*nextMove;
					alpha=val;
					hashf=hashfEXACT;
				}
				if (bestMove==EMPTY){
				   bestMove=*nextMove;
				}
			}//end only legal moves
		nextMove++;
		}//end all moves
	}//end all searchphases
    
    //*************************** End of search ********************************
	if (bestMove!=EMPTY){// !illegal_moves
       if (val!=INFINITE ||(val==INFINITE && alpha!=INFINITE)){
          hashtable1->recordHash(parent.boardkey, depth, alpha, hashf);
          hashtable1->storeBestMove(parent.boardkey, bestMove);
       }
       return alpha;
    }
    else{
       //no possible moves
       //current position in check?
       if ( parent.inCheck(parent.toMove) ){//checkmate
          hashtable1->recordHash(parent.boardkey, depth, -INFINITE, hashfEXACT);
          return -INFINITE;
       }
       else{//stalemate
          hashtable1->recordHash(parent.boardkey, depth, 0, hashfEXACT);
          return 0;
       }
    }      
}

int searchtree::quiescence(int move, board &parent, int alpha, int beta)
{
    nodes++;

	int val=hashtable1->probeHash(parent.boardkey, 0, alpha, beta);
    if (val!=NODATA){
       return val;
    } 

    int bestMove=EMPTY;
    int hashMove=EMPTY;
    int hashf=hashfALPHA;
    movelist ml;


    //*********************** Do an evaluation *****************************   
    //If this one is better we can return beta straight away.
    int eval=parent.evaluate(alpha, beta);
     
    if (eval>=beta){
          hashtable1->recordHash(parent.boardkey, 0, eval, hashfBETA);
          return beta;
    }
    if (eval>alpha){
          alpha=eval;
          hashf=hashfEXACT;
    }// */

   	for (int searchphase=0; searchphase<2; searchphase++){
		if (searchphase==0){
			//****************** Probe hashtable for best move *************************
			//We do this before generating moves.
			hashMove=hashtable1->probeBestMove(parent.boardkey);
			//is it a capturing move?
			if (parent.kind[(hashMove>>6)%64]==EMPTY){
			   //This move is NOT a capture move. Delete it and reset hashMove to EMPTY!
			   hashMove=EMPTY;
			}
			if(hashMove!=EMPTY){
				ml.moves[0]=hashMove;
				ml.moves[1]=0;
			}
			else{ml.moves[0]=0;}
		}
		else{
			//********************** Generate moves ***********************************
			//no move found in hash or it wasn't good enough. Generate moves...   
			parent.generateSortedCaptureMoves2(ml, kmt);
		}

		//********************** start real search ***********************************
		int* nextMove=&ml.moves[0];


		while (*nextMove!=0){
			    board child=parent;
			    child.makeMove(*nextMove);
				if(!child.isMoveIllegal(*nextMove)){//legal moves only
				val=-quiescence((*nextMove), child, -beta, -alpha);
					if (val>=INFINITE && alpha<INFINITE){ //move is winning move.
						//DONOT store INFINITE value in hash! This leads to great trouble 
						//if all moves are winning moves! It starts to play random moves 
						//then!
						return val;
					} 
					if (val>=beta){
						hashtable1->recordHash(parent.boardkey, 0, val, hashfBETA);
						//kmt.storeKillerMove(*nextMove, val, parent.ply, side(parent.toMove), parent.kind[((*nextMove)>>6)%64]);
						return beta;
					}
					if (val>alpha){
						bestMove=*nextMove;
						alpha=val;
						hashf=hashfEXACT;
					}
					if (bestMove==EMPTY){
						bestMove=*nextMove;
					}  
			  }   
			  nextMove++;
		}// */
	}


    hashtable1->recordHash(parent.boardkey, 0, alpha, hashf);      
    if (bestMove!=EMPTY)
       {hashtable1->storeBestMove(parent.boardkey, bestMove);}
    return alpha;      
}


searchtree::searchtree(hashtable &hashtableInput, EGTBProbe* initegtb){
    hashtable1=&hashtableInput;
    searchEnabled=&normalSearchtree;
	egtb=initegtb;
}

//this constructor is specifically for pondering searchtrees
searchtree::searchtree(hashtable &hashtableInput, int &enabledInit, EGTBProbe* initegtb){
    hashtable1=&hashtableInput;
    searchEnabled=&enabledInit;
	egtb=initegtb;
}

searchtree::~searchtree(){
}

void searchtree::printThinking(board b, MOVE firstMove){
	movelist ml;
	int moveNum=0;
	int bestMove=firstMove;

	while(bestMove!=EMPTY && !b.history1.doubleRepetition(b.absolPos, b.counter50move)){
		ml.moves[moveNum++]=bestMove;
		b.makeMove(bestMove);
		bestMove=hashtable1->probeBestMove(b.boardkey);
	}
	ml.moves[moveNum]=0;
	ml.moves[99]=0;//just to be sure

	ml.printMoveList();

}